#' Coerces input to image (array)
#' @inheritParams base::as.array
#' @param ... see \code{\link[base]{as.array}}
#' @family convert
#' @export
as.image <- as.array

#' Coerces input to image, shorthand for as.image
#' @param arr array to convert
#' @param transpose boolean, whether to transpose the resultant array.
#'
#'                 Note that Voir images are of dimension (nx,ny,nchanns),
#'                 and the returned array will also have dimensions
#'                 (nx,ny,nchanns).
#'              
#'                 That is, *the returned array will be transposed*, since
#'                  usually an image as an array would have dimensions
#'                  (ny,nx,nchanns), being (rows,columns,channels).
#'         
#'                 If \code{transpose=TRUE}, then the returned array will
#'                 have dimensions (ny, nx, nchanns).
#' @param ... IGNORED
#' @family convert
#' @export
array2im <- function(arr,transpose=FALSE,...)
{
    if ( transpose )
        arr <- imtranspose(arr)
    return( as.image(arr) )
}

#' Coerces image to array.
#' @inheritParams array2im
#' @param img image to convert
#' @family convert
#' @export
im2array <- function(img,transpose=FALSE,...)
{ 
    if (transpose)
        img <- imtranspose(img)
    return(as.array(img))
}

#' Checks whether the input is an image
#' @family convert
#' @export
is.image <- is.array


# #  NOTE: I like these versions better but if the image is 0 to 1 and you
# #  round it, you lose information.
# #  if you convert 0 to 255 then you can have unintended consequences when
# #  you further use the images.
# # ' Convert image to char (1-byte uint)
# # ' @param img image to convert
# # ' @return image as char (0-255).
# # ' @family convert
# # ' @export
# imchar <- function(img)
# {
#     return(roundAndCap(img,0,255))
# }
# 
# # ' Convert image to 4-byte int (-2^31+1 to 2^31)
# # ' @inheritParams imchar
# # ' @return image as char (-2^31+1 to 2^31)
# # ' @family convert
# # ' @export
# imint <- function(img)
# {
#     return(roundAndCap(img,-2^31+1,2^31))
# }

#' Convert image to char (DOESN'T DO ANYTHING)
#' @param img image to convert
#' @return input image.
#' 
#' @note These functions just return the input image as-is. 
#' They *could* scale the images to the appropriate range (0 to 1 maps to 0
#' to 255, e.g.) and round, but then one has to remember that the images no
#' longer range from 0 to 1 (which is the default for all the other
#' functions in this package).
#' @aliases imint
#' @rdname imchar
#' @family convert
#' @seealso \code{\link{imbinary}}, which *does* do converting.
#' @export
imchar <- function(img)
{
    return(img)
}
#' Convert image to 4-byte int (-2^31+1 to 2^31) (DOESN'T DO ANYTHING)
#' @inheritParams imchar
#' @family convert
#' @rdname imchar
#' @export
imint <- function(img)
{
    return(img)
}

#' Convert image to logical
#' @inheritParams imchar
#' @return image as binary (TRUE/FALSE)
#' @family convert
#' @export
imbinary <- function(img)
{
    return(as.logical(img))
}

#' converts to int and caps to range.
#' @keywords internal
#' @export
roundAndCap <- function(x, MIN, MAX, max=imageUtilities.option('max'))
{
    return(pmin(pmax(round(x),MIN),MAX))
}


#' gets number of pixels in the horizontal (x) direction
#' @param img image to get dimensions of
#' @param ... not used (backwards compatibility)
#' @return number of x pixels
#' @family dims
#' @export
imgetnx <- function(img,...)
{
    return(dim(img)[2])
}

#' gets number of pixels in the vertical (y) direction
#' @inheritParams imgetnx
#' @return number of y pixels
#' @family dims
#' @export
imgetny <- function(img,...)
{
    return(dim(img)[1])
}

#' gets dimensions of image, x, y, channs
#' @inheritParams imgetnx
#' @return vector (nx, ny, numchans)
#' @family dims
#' @export
imgetdims <- function(img)
{
    return( c( ncol(img), nrow(img), imgetnumcomp(img) ) )
}

#' gets number of image channels/components
#' @inheritParams imgetnx
#' @return number of channels.
#' @family dims
#' @export
imgetnumcomp <- function(img)
{
    return( ifelse(length(dim(img))==2,1,dim(img)[3]) )
}

#' downsamples an image
#' @param imgA   input image to be downsampled
#' @param dx     window size (horizontal dimension)
#' @param dy     window size (vertical dimension), default dx.
#' @param method one of "every" or "mean".
#'               If "every", every dx and dy'th pixel will be kept
#'                (starting from the first)
#'               If "mean", the mean of every non-overlapping dx by dy box
#'                will be calculated and that is retained.
#' @return downsampled image, of size ceiling(dim(imgA)/c(dy,dx)), where
#'         pixel (i,j) is calculated according to parameter "method".
#' @note \code{imsubsample} and \code{imdownsample} are aliases for the
#' same function.
#' @details 
#'      The \code{method="mean"} mode is \bold{very slow} at the moment. It
#'       takes ~11 user-seconds on a 5000x1000x3 image with dx=3, and
#'       ~28 user-seconds on a 5000x3000x3 image, dx=3.
#'      Consider whether you actually \emph{need} the mean in each cell
#'       before using this method; for display purposes
#'       \code{method="every"} does fine.
#'
#'      Also, \code{method="mean"} is implemented with a 0 border 
#'       condition (ie, if the dx by dy window goes outside of the image,
#'       missing values are assumed 0).
#' @aliases imsubsample imdownsample
#' @rdname imdownsample
#' @export
imdownsample <- function( imgA, dx, dy=dx, method=c('every','mean') )
{
    method <- match.arg(method,c('every','mean'))

    if ( method == "every" ) {
        return( imgA[ seq(1,nrow(imgA),by=dy), 
                      seq(1,ncol(imgA),by=dx), ] )
    } else if ( method == "mean" ) {
        sz <- dim(imgA)
        outSz <- c(ceiling(sz[1:2]/c(dy,dx)),sz[3])
        is <- pmin(seq(1,sz[1],by=dy) + floor(dy/2),sz[1]-1)
        js <- pmin(seq(1,sz[2],by=dx) + floor(dx/2),sz[2]-1)
        out3 <- array(NA,dim=outSz)
        for ( c in 1:sz[3] ) {
            out3[,,c] <- (t(filter(
                            t(filter(imgA[,,c],rep(1,dy))),
                            rep(1,dx))))[is,js]
        }
        out3 <- out3/(dx*dy)
        return(out3)
    }
}

#' @rdname imdownsample
#' @export
imsubsample <- imdownsample

#' combines single-channel images together to become multi-channel
#' @param ... input images to join, all of the same dimensions.
#' @return an image in which each of the input arguments forms a component
#' @export
imjoin <- function(...)
{
    ims <- list(...)
    if ( length(ims) == 1 )
        return(ims[[1]])

    if ( !all(sapply(ims,is.image)) )
        stop("imjoin: all inputs must be images!")
    # drop singletons.
    ims <- lapply(ims,drop)
    sz <- dim(ims[[1]])
    if ( length(sz) > 2 )
        stop("imjoin: all inputs must be single-channel!")
    if ( is.list((dims <- sapply(ims,dim))) ||
         !all(sz==dims) )
        stop("imjoin: all inputs must have same dimensions!")

    outI <- array( unlist(ims),dim=c(sz,length(ims)) )
}

#' Produces a subset of an image
#' @param img input image to subset
#' @param fx  first x/col (1-based)
#' @param fy  first y/row (1-based)
#' @param lx  last x/col (1-based)
#' @param ly  last y/row (1-based)
#' @param one.based boolean, default TRUE, whether fx, fy, lx and ly are
#'                  one or zero based.
#' @return a (ly-fy+1) by (lx-fx+1) by (numchanns) array containing the
#' subset of the image from fx to lx, fy to ly inclusive.
#' @note This should mask Voir's imsubset, however, it may not.
#' @export
imsubset <- function( img, fx, fy, lx, ly, one.based=TRUE )
{
    if ( !one.based ) {
        fx <- fx+1
        fy <- fy+1
        lx <- lx+1
        ly <- ly+1
    }
    if ( length(dim(img))==2 )
        return( img[fy:ly,fx:lx] )
    else
        return( img[fy:ly,fx:lx,] )
}

#' pastes patch image into src image starting at (src) fx,fy
#' @param src source image
#' @param patch image to paste into src, of same number of channels as src.
#' @param fx start x/row of src to paste patch into.
#' @param fy start y/col of src to paste patch into.
#' @param one.based whether fx and fy are one-based or 0-based.
#' @return the image src with patch pasted in at fx, fy
#' @export
impaste <- function( patch, src, fx, fy, one.based=TRUE ) 
{
    if ( (length(dim(src))!=length(dim(patch))) ||
         (length(dim(src))==3 && dim(src)[3]!=dim(patch)[3]) ) 
        stop("impaste: src and patch images should have same number of channels!")

    if ( !one.based ) { 
        fx <- fx + 1 
        fy <- fy + 1 
    } 

    srcI <- src
    # argh indexing annoyances with single-channel.
    if ( length(dim(srcI))==2 ) 
        dim(srcI)<-c(dim(srcI),1)
    sz   <- dim(patch)

    srcI[ fy:(fy+sz[1]-1), fx:(fx+sz[2]-1), ] <- patch

    return(srcI)
}
